/*
 * Copyright (c) 2002 Falk Feddersen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


/*    
      bubble.h
*/


#ifndef BUBBLE_H
#define BUBBLE_H

#include "defs.h"

typedef struct {
  switch_t bubble_flag;
  switch_t ic_flag;
  breaking_t bubble_source;
  double dr;
  int nr;
  double rmin, rmax;
  double beta0;
  double scale_h;
} bubbleinfo;


//void terminal_velocity(field1D *wt, field1D *r);

//void init_bubble_stuff(int n, int zlevel, double rmin, double rmax, double beta0, double scale_h);
void init_bubble_stuff(int zlevel, bubbleinfo *B);

void deallocate_bubble_fields();

double calculate_void_fraction_point(field2D *NB, int zlevel);

double calculate_void_fraction0(field1D *NR);

void calculate_void_fraction(const field2D *NB, field1D *BETA);

void calculate_bubble_breakup_energy_loss(field2D *NB, field1D *LOSS);

void calculate_buoyancy_flux(field1D *BFLUX, const field2D *NB);

/* Bubble buoyancy flux, breakup, dissolution terms */

void   calculate_bubble_stats(field1D *DISS);


/* Bubble boundary condition functions */

void update_bubble_bc0(field2D *NN);
void update_bubble_bc(field2D *NN, field1D *Kv);

/* Bubble dynamics */

void rhs_bubble_eq(field2D *NBT, field1D *BFLUX, const field2D *NB, const field1D *Kv);

void rhs_implicit_bubble_eq(field2D *NBT, const field2D *NB);

#endif /* BUBBLE_H */
