

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <glib.h>

#include "field.h"

#define __G_LIB_H_

/* function definitions */

field1D*  allocate_field1D(int N)
{
  field1D *T;

#ifdef __G_LIB_H_   /* if glib is defined */
  T = (field1D *) g_malloc(sizeof(field1D));
  T->data = (double *) g_malloc(N*sizeof(double));
#else
  T = (field1D *) malloc(sizeof(field1D));
  T->data = (double *) malloc(N*sizeof(double));
#endif  /* GLIB */

  T->N = N;

  return T;
}

field2D*  allocate_field2D(int N,int M)
{
  int i;
  field2D *T;

#ifdef __G_LIB_H_   /* if glib is defined */
  T = (field2D *) g_malloc(sizeof(field2D));  
  T->data = (double *) g_malloc(N*M*sizeof(double));
  T->row = (double **) g_malloc(N*sizeof(void*));
#else
  T = (field2D *) malloc(sizeof(field2D));  
  T->data = (double *) malloc(N*M*sizeof(double));
  T->row = (double **) malloc(N*sizeof(void*));
#endif

  T->N = N;
  T->M = M;

  T->num = N*M;


  for (i=0;i<N;i++) 
    T->row[i] = &(T->data[i*M]);

  return T;
}

void deallocate_field1D(field1D *T)
{


  if (T->data!=NULL) {
#ifdef __G_LIB_H_   /* if glib is defined */
    g_free(T->data);
#else
    free(T->data);
#endif /* GLIB */
  }

  if (T!=NULL) {
#ifdef __G_LIB_H_   /* if glib is defined */
    g_free(T);
#else
    free(T);
#endif /* GLIB */
   }
}


void deallocate_field2D(field2D *T)
{

  if (T->row!=NULL)
#ifdef __G_LIB_H_   /* if glib is defined */
    g_free(T->row);
#else
    free(T->row);
#endif /* GLIB */


  if (T->data!=NULL) {
#ifdef __G_LIB_H_   /* if glib is defined */
    g_free(T->data);
#else
    free(T->data);
#endif /* GLIB */
  }

  if (T!=NULL) {
#ifdef __G_LIB_H_   /* if glib is defined */
    g_free(T);
#else
    free(T);
#endif /* GLIB */
   }

}



int  set_field1D_to_value(field1D *T, double a)
{
  int i;

  if ((T==NULL)||(T->data==NULL))
    return 0;

  for (i=0;i<T->N;i++)
    T->data[i] = a;

  return 1;
}

/* Copies field1D S into T -> must be allocated */

int  copy_field1D(field1D *T, const field1D *S)
{
  int i;

  if ((T==NULL)||(T->data==NULL))
    return 0;

  if ((S==NULL)||(S->data==NULL))
    return 0;

  if (T->N != S->N) 
    return 0;

  for (i=0;i<S->N;i++)
    T->data[i] = S->data[i];

  return 1;
}


int copy_field1D_mult(field1D *T, const field1D *S, double a)
{
  int i;

  if ((T==NULL)||(T->data==NULL))
    return 0;

  if ((S==NULL)||(S->data==NULL))
    return 0;

  if (T->N != S->N) 
    return 0;

  for (i=0;i<S->N;i++)
    T->data[i] = S->data[i] * a;

  return 1;
}


int mult_field1D(field1D *T, const field1D *S)
{
  int i;

  if ((T==NULL)||(T->data==NULL))
    return 0;

  if ((S==NULL)||(S->data==NULL))
    return 0;

  if (T->N != S->N) 
    return 0;

  for (i=0;i<S->N;i++)
    T->data[i] *= S->data[i];

  return 1;
}


void mult_field1D_const(field1D *T, double a)
{
  int i;

  for (i=0;i<T->N;i++)
    T->data[i] *= a;

}

int  set_field2D_to_value(field2D *T, double a)
{
  int i;

  if ((T==NULL)||(T->data==NULL))
    return 0;

  for (i=0;i<T->num;i++)
    T->data[i] = a;

  return 1;
}


int copy_field2D_mult(field2D *T, const field2D *S, double a)   // T = a * S 
{
  int i;

  if ((T==NULL)||(T->data==NULL))
    return 0;

  if ((S==NULL)||(S->data==NULL))
    return 0;

  if (T->N != S->N) 
    return 0;

  for (i=0;i<S->num;i++)
    T->data[i] = S->data[i] * a;

  return 1;
}


int copy_field2D(field2D *T, const field2D *S)   // T = a * S 
{
  int i;

  if ((T==NULL)||(T->data==NULL))
    return 0;

  if ((S==NULL)||(S->data==NULL))
    return 0;

  if (T->N != S->N) 
    return 0;

  for (i=0;i<S->num;i++)
    T->data[i] = S->data[i];

  return 1;
}

void mult_field2D_const(field2D *T, double a)
{
  int i;

  for (i=0;i<T->num;i++)
    T->data[i] *= a;

}


double max_field2D(field2D *T)
{
  int i;
  double tmp,max=-1.0;

  for (i=0;i<T->num;i++) {
    tmp = fabs(T->data[i]);
    if (tmp>max) 
      max = tmp;
  }

  return max;
}


double max_field1D(field1D *T)
{
  int i;
  double tmp,max=-1.0;

  for (i=0;i<T->N;i++) {
    tmp = fabs(T->data[i]);
    if (tmp>max) 
      max = tmp;
  }

  return max;
}


int max_field1D_location(field1D *T)
{
  int i;
  double tmp,max=-1.0;
  int loc = -1;

  for (i=0;i<T->N;i++) {
    tmp = fabs(T->data[i]);
    if (tmp>max) {
      max = tmp;
      loc = i;
    }
  }

  return loc;
}

/* File handling functions */


FILE*  open_file_ptr_for_writing(char *s)
{
  FILE *fp;

  if ((fp=fopen(s,"w"))==NULL) {
    fprintf(stderr,"Cannot open file for writing: %s\n",s);
    exit(-1);
  }

  return fp;
}

FILE*  open_file_ptr_for_reading(char *s)
{
  FILE *fp;

  if ((fp=fopen(s,"r"))==NULL) {
    fprintf(stderr,"Cannot open file for reading: %s\n",s);
    exit(-1);
  }

  return fp;
}

/* field1d Input functions */


int load_field1D_ascii(char *fname, field1D *T)    /* returns 1 if successfull */
{
   FILE *fp;
   int i,numc,rval=1;
   double a;

   fp = open_file_ptr_for_reading(fname);

   for (i=0;i<T->N;i++) {
     numc = fscanf(fp,"%lf",&a);
     if ((numc == 0)||(numc==EOF))
       rval = 0;
     DR1(T,i) = a;
   }

   fclose(fp);
   return rval;
}

int load_field1D_binary(char *fname, field1D *T)    /* returns 1 if successfull */
{
   FILE *fp;
   int rval=1;
   size_t sized,numt,count2;

   fp = open_file_ptr_for_reading(fname);

   numt = (size_t ) T->N;

   sized = sizeof(double);
   count2 = fread(T->data,sized,numt,fp);

   if (numt != count2)
     rval=0;

   fclose(fp);
   return rval; 
}

/* field1d Output functions */


void save_field1D_ascii(char *fname, field1D *T)
{
  FILE *fp;
  int i;

  fp = open_file_ptr_for_writing(fname);

  for (i=0;i<T->N;i++) {
    fprintf(fp,"%g\n",DR1(T,i));
  }
  fprintf(fp,"\n");


  fclose(fp);

}


void save_field1D_binary(char *fname, field1D *T)
{
  FILE *fp;
  int size;

  fp = open_file_ptr_for_writing(fname);

  size = sizeof(double);

  fwrite(T->data,size,T->N,fp);

  fclose(fp);

}


/* field2d Input functions */


int load_field2D_ascii(char *fname, field2D *T)    /* returns 1 if successfull */
{
   FILE *fp;
   int i,j,numc,rval=1;
   double a;

   fp = open_file_ptr_for_reading(fname);

   for (i=0;i<T->N;i++) {
     for (j=0;j<T->M;j++) {
        numc = fscanf(fp,"%lf",&a);
        if ((numc == 0)||(numc==EOF)) {
	  rval = 0;
          return rval;
	}
        DR2(T,i,j) = a;
     }
   }
   fclose(fp);
   return rval;
}

int load_field2D_binary(char *fname, field2D *T)    /* returns 1 if successfull */
{
   FILE *fp;
   int rval=1;
   size_t sized,numt,count2;

   fp = open_file_ptr_for_reading(fname);

   numt = (size_t ) T->num;

   sized = sizeof(double);
   count2 = fread(T->data,sized,numt,fp);

   if (numt != count2)
     rval=0;

   fclose(fp);
   return rval; 
}

/* field2d Output functions */


void print_field2D(field2D *T)
{
  //  FILE *fp;
  int i,j;

  for (i=0;i<T->N;i++) {
    for (j=0;j<T->M;j++) 
      fprintf(stdout,"%g ",DR2(T,i,j));
    fprintf(stdout,"\n");
  }

}


void save_field2D_ascii(char *fname, field2D *T)
{
  FILE *fp;
  int i,j;

  fp = open_file_ptr_for_writing(fname);

  for (i=0;i<T->N;i++) {
    for (j=0;j<T->M;j++) 
      fprintf(fp,"%g ",DR2(T,i,j));
    fprintf(fp,"\n");
  }

  fclose(fp);

}


void save_field2D_binary(char *fname, field2D *T)
{
  FILE *fp;
  int size;

  fp = open_file_ptr_for_writing(fname);

  size = sizeof(double);

  fwrite(T->data,size,T->num,fp);

  fclose(fp);

}


double max_array(double *arr, int num)
{
  int i,index=-1;
  double tmp,max=-1.0;

  for (i=0;i<num;i++) {
    tmp = fabs(arr[i]);
    if (tmp>max) {
      index = i; 
      max = tmp;
    }
  }

  printf("Max array = %g  at location %d\n",max,index);
  
  return max;
}

