/*
 * Copyright (c) 2002 Falk Feddersen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/* 
   implicit.c 


*/

#include<math.h>
#include<glib.h>

#include "defs.h"
#include "implicit.h"
#include "tri.h"

extern double ustar;
extern double vstar;

/* These are the diagonal, upper, & lower part of tridiagonal matrices */

double *dd;
double *ddu;
double *ddv;
double *ddb;
double *dde;
double *udd;
double *ldd;
double *udb;
double *ldb;
double *ude;
double *lde;
double *rhsu;   /* rhs of U mean flow equation */
double *rhsv;   /* rhs of V mean flow equation */
double *rhsb;   /* rhs of tke equation     */
double *rhse;   /* rhs of epsilon equation */
double *rhss;   /* rhs of sediment equation */

field2D *RHS_BUBBLE=NULL;   /* rhs of the bubble equation */
field2D *XX_BUBBLE=NULL;    /* solution of tridiagonal solver */

double *work;   // extra room to do computations
double *work2;

void init_implicit_solver(int nz, int nr, switch_t bubble_flag)
{
  int i;

  dd = (double * ) g_malloc(sizeof(double)*(nz-2));
  ddu = (double * ) g_malloc(sizeof(double)*(nz-2));
  ddv = (double * ) g_malloc(sizeof(double)*(nz-2));
  ddb = (double * ) g_malloc(sizeof(double)*(nz-2));
  dde = (double * ) g_malloc(sizeof(double)*(nz-2));
  dd = (double * ) g_malloc(sizeof(double)*(nz-2));
  udd = (double * ) g_malloc(sizeof(double)*(nz-2));
  ldd = (double * ) g_malloc(sizeof(double)*(nz-2));
  udb = (double * ) g_malloc(sizeof(double)*(nz-2));
  ldb = (double * ) g_malloc(sizeof(double)*(nz-2));
  ude = (double * ) g_malloc(sizeof(double)*(nz-2));
  lde = (double * ) g_malloc(sizeof(double)*(nz-2));
  work = (double * ) g_malloc(sizeof(double)*(nz-2));
  work2 = (double * ) g_malloc(sizeof(double)*(nz-2));
  rhsu = (double * ) g_malloc(sizeof(double)*(nz-2));
  rhsv = (double * ) g_malloc(sizeof(double)*(nz-2));
  rhsb = (double * ) g_malloc(sizeof(double)*(nz-2));
  rhse = (double * ) g_malloc(sizeof(double)*(nz-2));
  rhss = (double * ) g_malloc(sizeof(double)*(nz-2));

  if (bubble_flag == ON) {    // allocate bubble stuff 

    RHS_BUBBLE = allocate_field2D(nz-2,nr);   /* rhs of  A x = b */

  /* This stuff is taken from field.c:  we want XX_BUBBLE to be part of NB */

    XX_BUBBLE = (field2D *) g_malloc(sizeof(field2D));
    XX_BUBBLE->row = (double **) g_malloc((nz-2)*sizeof(void*));
    XX_BUBBLE->N = nz-2;
    XX_BUBBLE->M = nr;
    XX_BUBBLE->data = NB->data + nr;

    for (i=0;i<XX_BUBBLE->N;i++) 
      XX_BUBBLE->row[i] = &(XX_BUBBLE->data[i*(XX_BUBBLE->M)]);

  }

}

void deallocate_implicit_solver()
{

  g_free(dd);
  g_free(ddu);
  g_free(ddv);
  g_free(ddb);
  g_free(dde);
  g_free(udd);
  g_free(ldd);
  g_free(ude);
  g_free(lde);
  g_free(udb);
  g_free(ldb);
  g_free(work);
  g_free(work2);
  g_free(rhsu);
  g_free(rhsv);
  g_free(rhsb);
  g_free(rhse);
  g_free(rhss);

  if (RHS_BUBBLE!=NULL) {
    deallocate_field2D(RHS_BUBBLE);
  
    g_free(XX_BUBBLE->row);
    g_free(XX_BUBBLE);
  }


}



void implicit_solve_mflow_euler(field1D *U, const field1D *Ut, field1D *V, 
                                const field1D *Vt, const field1D *Kv, 
                                double dt)
{

  int i, ok;
  int N = U->N;

  double *vd = V->data+1;
  double *ud = U->data+1;
  double vv, vp, uu, up;

  const double *utd = Ut->data;
  const double *vtd = Vt->data;
  
  double *uwd = UW->data;
  double *vwd = VW->data;

  const double *kvd = Kv->data;

  double kv,kvm;
  double Fup, Fum, Fvp, Fvm, tmp;
  double vdiffusion, udiffusion;

  double hdt = 0.5*dt;

  /* First calculate the RHS stuff */
  /* do the first grid minus flux before looping */

  uu = *ud++;
  vv = *vd++;   

  kvm = *kvd++;
  Fum = kvm * (uu - DR1(U,0));
  Fvm = kvm * (vv - DR1(V,0));

  *uwd++ = idz*Fum;
  *vwd++ = idz*Fvm;

  for (i=1;i<N-1;i++) {

    vp = *vd++;
    up = *ud++;

    Fup = (*kvd) * ( up - uu);
    Fvp = (*kvd++) * (vp - vv);

    *uwd++ = idz*Fup;     // save the diffusive fluxes
    *vwd++ = idz*Fvp;

    //    Fvp = (*kv++) * (*vdp++ - vv);
    //    Fum = (*kvm) * (uu - *udm++);
    //    Fvm = (*kvm++) * (vv - *vdm++);

    udiffusion = idz2*(Fup-Fum);
    vdiffusion = idz2*(Fvp-Fvm);
    rhsu[i-1] = uu + dt*(0.5*udiffusion + *utd++);
    rhsv[i-1] = vv + dt*(0.5*vdiffusion + *vtd++);

    Fum = Fup;
    Fvm = Fvp;

    uu = up;
    vv = vp;
  }   /* for (i=0.... */

  /* Now the tridiagonal matrix is set up */
  /* i = 0 -> here we apply no-slip bottom bc */

  kvd = Kv->data;

  kvm = *kvd++;
  kv = *kvd++;

  ddv[0] = 1 + hdt*idz2*(kv+kvm);
  udd[0] = -hdt*kv*idz2;
  ldd[0] = 0.0;
  kvm = kv;

  /* i=1 to < N-3 */

  for (i=1;i<N-3;i++) {
    kv = *kvd++;
    ddv[i] = 1 + hdt*idz2*(kvm+kv);
    udd[i] = -hdt*kv*idz2;
    ldd[i] = -hdt*kvm*idz2;
    kvm = kv;
  }

  /* i = N-3, here deal with top flux BC */

  i = N-3;
  kv = *kvd++;
  ldd[i] = -hdt*kvm*idz2;
  udd[i] = 0.0;
  ddv[i] = 1.0 + hdt*idz2*(kvm);

  /* Now handle the top boundary condition -> ustar & vstar to the rhs */

  rhsu[i] += idz*hdt*ustar*ustar;
  rhsv[i] += idz*hdt*vstar*vstar;


  /* Now do the implicit solving */
  /* First for U then for V */

  ud = U->data+1;   // reset the ud pointer
  vd = V->data+1;

  ok = tridiagonal_mflow(ddv, ldd, udd, rhsu, rhsv, ud, vd, work, N-2);
  if (ok != 0)
    error_message("** Critical Error:  Tridiagonal call for mean flow failed!");

  /* Next is U & V boundary conditions */
  /* Bottom BC */

  DR1(U,0) = 0.0;
  DR1(V,0) = 0.0;   // hah!  these are easy

  /* Surface BC */
  kv = DR1(Kv,N-2);
  DR1(U,N-1) = DR1(U,N-2) + SQR(ustar)*dz/kv;
  DR1(V,N-1) = DR1(V,N-2) + SQR(vstar)*dz/kv;


}



void implicit_solve_mflowAB2(field1D *U, const field1D *Ut1, const field1D *Ut2,
                             field1D *V, const field1D *Vt1, const field1D *Vt2,
                             const field1D *Kv, double dt)
{

  int i, ok;
  int N = U->N;

  double *vd = V->data+1;
  double *ud = U->data+1;
  double vv, vp, uu, up;

  const double *ut1d = Ut1->data;
  const double *vt1d = Vt1->data;
  const double *ut2d = Ut2->data;
  const double *vt2d = Vt2->data;
  
  double *uwd = UW->data;
  double *vwd = VW->data;

  const double *kvd = Kv->data;

  double kv,kvm;
  double Fup, Fum, Fvp, Fvm, tmp;
  double vdiffusion, udiffusion;

  double hdt = 0.5*dt;

  /* First calculate the RHS stuff */
  /* do the first grid minus flux before looping */

  kvm = *kvd++;

  vv = *vd++;
  uu = *ud++;

  Fum = kvm * (uu - DR1(U,0));
  Fvm = kvm * (vv - DR1(V,0));

  *uwd++ = idz*Fum;
  *vwd++ = idz*Fvm;

  for (i=1;i<N-1;i++) {

    vp = *vd++;
    up = *ud++;

    Fup = (*kvd) * ( up - uu);
    Fvp = (*kvd++) * (vp - vv);

    *uwd++ = idz*Fup;
    *vwd++ = idz*Fvp;

    //    Fvp = (*kv++) * (*vdp++ - vv);
    //    Fum = (*kvm) * (uu - *udm++);
    //    Fvm = (*kvm++) * (vv - *vdm++);

    udiffusion = idz2*(Fup-Fum);
    vdiffusion = idz2*(Fvp-Fvm);
    rhsu[i-1] = uu + hdt*(udiffusion + 3.0 * (*ut1d++) - (*ut2d++));
    rhsv[i-1] = vv + hdt*(vdiffusion + 3.0 * (*vt1d++) - (*vt2d++));

    Fum = Fup;
    Fvm = Fvp;

    uu = up;
    vv = vp;

  }

  /* i = 0 -> here we apply no-slip bottom bc */

  
  kvd = Kv->data;

  kvm = *kvd++;
  kv = *kvd++;

  ddv[0] = 1 + hdt*idz2*(kv+kvm);
  udd[0] = -hdt*kv*idz2;
  ldd[0] = 0.0;
  kvm = kv;

  /* i=1 to < N-3 */

  for (i=1;i<N-3;i++) {
    kv = *kvd++;
    ddv[i] = 1 + hdt*idz2*(kvm+kv);
    udd[i] = -hdt*kv*idz2;
    ldd[i] = -hdt*kvm*idz2;
    kvm = kv;
  }

  /* i = N-3, here deal with top flux BC */

  i = N-3;

  ldd[i] = -hdt*kvm*idz2;
  udd[i] = 0.0;
  ddv[i] = 1.0 + hdt*idz2*(kvm);

  /* Now handle the top boundary condition -> ustar & vstar to the rhs */

  rhsu[i] += idz*hdt*ustar*ustar;
  rhsv[i] += idz*hdt*vstar*vstar;


  /* Now do the implicit solving for both U & V */

  ud = U->data+1;   // reset the ud pointer
  vd = V->data+1;

  ok = tridiagonal_mflow(ddv, ldd, udd, rhsu, rhsv, ud, vd, work, N-2);
  if (ok != 0)
    error_message("** Critical Error:  Tridiagonal call for mean flow failed!");

  /* Next is U & V boundary conditions */
  /* Bottom BC */
  DR1(U,0) = 0.0;
  DR1(V,0) = 0.0;   // hah!  these are easy

  /* Surface BC */
  kv = DR1(Kv,N-2);
  DR1(U,N-1) = DR1(U,N-2) + SQR(ustar)*dz/kv;
  DR1(V,N-1) = DR1(V,N-2) + SQR(vstar)*dz/kv;


}




void implicit_solve_tke_euler(field1D *B, const field1D *Bt, const field1D *Kv, 
                              double k_bed, double dt)
{

  int i, ok;
  int N = B->N;
  double hdt = 0.5*dt;

  double *bd = B->data+1;
  double bb, bp;
  
  const double *btd = Bt->data;
  
  const double *kvd = Kv->data;

  double kv,kvm;
  double Fbp, Fbm, tmp;
  double bdiffusion;

  /* First calculate the RHS stuff */
  /* do the first grid minus flux before looping */

  bb = *bd++;
  kvm = *kvd++;
  Fbm = kvm * (bb - DR1(B,0));

  for (i=1;i<N-1;i++) {

    bp = *bd++;

    Fbp = (*kvd++) * ( bp - bb);

    bdiffusion = idz2*(Fbp-Fbm);
    rhsb[i-1] = bb + dt*(0.5*bdiffusion + *btd++);

    Fbm = Fbp;
    bb = bp;

  }

  /* i = 0 -> here we set prod=diss bottom bc */

  kvd = Kv->data;

  kvm = *kvd++;
  kv = *kvd++;
  ddb[0] = 1 + hdt*idz2*(kv+kvm);
  udd[0] = -hdt*kv*idz2;
  ldd[0] = 0.0;

  /* What goes on rhs of rhb[0] for this??? */

  rhsb[0] += hdt*idz2*kvm*k_bed;

  kvm = kv;

  /* i=1 to < N-3 */

  for (i=1;i<N-3;i++) {
    kv = *kvd++;
    tmp = 1 + hdt*idz2*(kvm+kv);
    ddb[i] = tmp;
    udd[i] = -hdt*kv*idz2;
    ldd[i] = -hdt*kvm*idz2;
    kvm = kv;
  }

  /* i = N-3, here deal with top flux BC */

  i = N-3;
  kv = *kvd++;
  ldd[i] = -hdt*kvm*idz2;
  udd[i] = 0.0;
  ddb[i] = 1.0 + hdt*idz2*(kvm);

  /* Now handle the top boundary condition -> ustar & vstar to the rhs */

  rhsb[i] += idz*hdt*(wave_tke_surface_flux);
  //    error_message("This stuff doesn't work!!!");

  /* Now do the implicit solving */
  /* First for U then for V */

  bd = B->data+1;   // reset the ud pointer

  ok = tridiagonal(ddb, ldd, udd, rhsb, bd, work, N-2);
  if (ok != 0)
    error_message("** Critical Error:  Tridiagonal call for TKE failed!");


  /* Next is tke boundary conditions */
  /* Bottom BC */
  DR1(B,0) = k_bed;

  /* Surface BC */
  kv = DR1(Kv,N-2);
  DR1(B,N-1) = DR1(B,N-2) + wave_tke_surface_flux*dz/kv;

}



void implicit_solve_tkeAB2(field1D *B, const field1D *Bt1, const field1D *Bt2, const field1D *Kv, 
                        double k_bed, double dt)
{

  int i, ok;
  int N = B->N;
  double hdt = 0.5*dt;

  double *bd = B->data+1;
  double bb, bp;
  
  const double *bt1d = Bt1->data;
  const double *bt2d = Bt2->data;
  
  const double *kvd = Kv->data;

  double kv,kvm;
  double Fbp, Fbm, tmp;
  double bdiffusion;

  /* First calculate the RHS stuff */
  /* do the first grid minus flux before looping */

  bb = *bd++;
  kvm = *kvd++;
  Fbm = kvm * (bb - DR1(B,0));

  for (i=1;i<N-1;i++) {

    bp = *bd++;

    Fbp = (*kvd++) * ( bp - bb);

    bdiffusion = idz2*(Fbp-Fbm);
    rhsb[i-1] = bb + hdt*(bdiffusion + 3.0 * (*bt1d++) - (*bt2d++));

    Fbm = Fbp;
    bb = bp;

  }

  /* i = 0 -> here we set prod=diss bottom bc */

  kvd = Kv->data;

  kvm = *kvd++;
  kv = *kvd++;
  ddb[0] = 1 + hdt*idz2*(kv+kvm);
  udd[0] = -hdt*kv*idz2;
  ldd[0] = 0.0;

  /* What goes on rhs of rhb[0] for this??? */

  rhsb[0] += hdt*idz2*kvm*k_bed;

  kvm = kv;

  /* i=1 to < N-3 */

  for (i=1;i<N-3;i++) {
    kv = *kvd++;
    tmp = 1 + hdt*idz2*(kvm+kv);
    ddb[i] = tmp;
    udd[i] = -hdt*kv*idz2;
    ldd[i] = -hdt*kvm*idz2;
    kvm = kv;
  }

  /* i = N-3, here deal with top flux BC */

  i = N-3;
  kv = *kvd++;
  ldd[i] = -hdt*kvm*idz2;
  udd[i] = 0.0;
  ddb[i] = 1.0 + hdt*idz2*(kvm);

  /* Now handle the top boundary condition -> ustar & vstar to the rhs */

  rhsb[i] += idz*hdt*(wave_tke_surface_flux);
  //    error_message("This stuff doesn't work!!!");

  /* Now do the implicit solving */
  /* First for U then for V */

  bd = B->data+1;   // reset the ud pointer

  ok = tridiagonal(ddb, ldd, udd, rhsb, bd, work, N-2);
  if (ok != 0)
    error_message("** Critical Error:  Tridiagonal call for TKE failed!");


  /* Next is tke boundary conditions */
  /* Bottom BC */
  DR1(B,0) = k_bed;

  /* Surface BC */
  kv = DR1(Kv,N-2);
  DR1(B,N-1) = DR1(B,N-2) + wave_tke_surface_flux*dz/kv;

}

/* ------- checking routines ---------- */

void check_dissipation_stratified(field1D *DISS, field1D *N2);

void check_dissipation(field1D *DISS) 
{

  const static double diss0 = 1e-11;

  int N = DISS->N;
  int i;

  double *dissd = DISS->data;

  for (i=1;i<N;i++) {
    if (*dissd < diss0)
      *dissd = diss0;
    dissd++;
  }

}

void check_tke(field1D *TKE) 
{

  const static double tke0 = 1e-7;

  int N = TKE->N;
  int i;

  double *tked = TKE->data;

  for (i=0;i<N;i++) {
    if (*tked < tke0)
      *tked = tke0;
    tked++;
  }

}





/* --------------- kepsilon implicit solvers --------------- */

#define EPSILON_DIRCHELET

void implicit_solve_kepsilon_euler(field1D *B, const field1D *Bt, field1D *DISS, const field1D *DISSt,
                              const field1D *Kv, double k_bed, double diss_bed, double dt)
{

  int i, ok;
  int N = B->N;
  double hdt = 0.5*dt;

  static const double Sige =   1.077498753010693;
  static const double iSige =   9.280753200000001e-01;   //   this is 1/\sigma_\epsilon in GOTM
  static double cube_Cmu0 =   1.720651643280000e-01;

  double *bd = B->data+1;
  double *dissd = DISS->data+1;

  double bb, bp;
  double dd, dp;
  double qq;
  
  const double *btd = Bt->data;
  const double *dtd = DISSt->data;  
  const double *kvd = Kv->data;

  double kv,kvm;
  double Ftke_p, Ftke_m, Fepsilon_m, Fepsilon_p, tmp;
  double tke_diffusion, ediffusion;
  double diss_top;

  /* First calculate the RHS stuff */
  /* do the first grid minus flux before looping */

  bb = *bd++;
  dd = *dissd++;
  kvm = *kvd++;
  Ftke_m = kvm * (bb - DR1(B,0));
  Fepsilon_m = iSige * kvm * (dd - DR1(DISS,0));

  for (i=1;i<N-1;i++) {

    bp = *bd++;
    dp = *dissd++;
    kv = *kvd++;

    Ftke_p = kv * ( bp - bb);
    Fepsilon_p = iSige * kv * (dp - dd);
    tke_diffusion = idz2*(Ftke_p-Ftke_m);
    ediffusion = idz2*(Fepsilon_p - Fepsilon_m);
    
    rhsb[i-1] = bb + dt*(0.5*tke_diffusion + *btd++);

    rhse[i-1] = dd + dt*(0.5*ediffusion + *dtd++);

    Ftke_m = Ftke_p;
    Fepsilon_m = Fepsilon_p;
    bb = bp;
    dd = dp;

  }

  /* i = 0 -> here we set prod=diss bottom bc */

  kvd = Kv->data;

  kvm = *kvd++;
  kv = *kvd++;
  ddb[0] = 1 + hdt*idz2*(kv+kvm);
  udb[0] = -hdt*kv*idz2;
  ldb[0] = 0.0;
  dde[0] = 1 + iSige*hdt*idz2*(kv+kvm);
  ude[0] = -hdt*iSige*kv*idz2;
  lde[0] = 0.0;


  /* What goes on rhs of rhb[0] for this??? */

  rhsb[0] += hdt*idz2*kvm*k_bed;
  rhse[0] += hdt*idz2*iSige*kvm*diss_bed; 

  kvm = kv;

  /* i=1 to < N-3 */

  for (i=1;i<N-3;i++) {
    kv = *kvd++;
    tmp = hdt*idz2*(kvm+kv);
    ddb[i] = 1 + tmp;
    dde[i] = 1 + iSige*tmp;
    udb[i] = -hdt*kv*idz2;
    ude[i] = -iSige*hdt*kv*idz2;
    ldb[i] = -hdt*kvm*idz2;
    lde[i] = -iSige*hdt*kvm*idz2;
    kvm = kv;
  }

  /* i = N-3, here deal with top flux BC */

  i = N-3;
  kv = *kvd++;
  ldb[i] = -hdt*kvm*idz2;
  udb[i] = 0.0;
  ddb[i] = 1.0 + hdt*idz2*(kvm);
  lde[i] = -iSige*hdt*kvm*idz2;
  ude[i] = 0.0;

#ifdef EPSILON_DIRCHELET
  dde[i] = 1.0 + iSige*hdt*idz2*(kvm+kv);   // this is a top Dirchelet BC!
#endif
#ifdef EPSILON_FLUX
  dde[i] = 1.0 + iSige*hdt*idz2*(kvm);   // this is a top flux BC! a la Burchard
#endif

  /* Now handle the top tke boundary condition -> surface wave flux to rhs */

  rhsb[i] += idz*hdt*(wave_tke_surface_flux);

  /* Now do the implicit solving */
  /* First for U then for V */

  bd = B->data+1;   // reset the tke pointer


  ok = tridiagonal(ddb, ldb, udb, rhsb, bd, work, N-2);
  if (ok != 0)
    error_message("** Critical Error:  Tridiagonal call for k (in k-e) failed!");


  /* Next is tke boundary conditions */
  /* Bottom BC */
  DR1(B,0) = k_bed;

  /* Surface BC */
  kv = DR1(Kv,N-2);
  DR1(B,N-1) = DR1(B,N-2) + wave_tke_surface_flux*dz/kv;

  check_tke(B);

  /* figure out what the top dissipation should be */

  /* Next to the epsilon part */

  dissd = DISS->data+1;

  qq = sqrt(DR1(B,N-1));

#ifdef EPSILON_DIRCHELET
  diss_top  = cube_Cmu0 * CUBE(qq) / DR1(LL,N-1);
  rhse[N-3] += idz2*hdt*kv*diss_top;
#endif
#ifdef EPSILON_FLUX
  diss_top  = -cube_Cmu0 * kv * iSige * CUBE(qq) / (DR1(LL,N-1) * 0.2);
  rhse[N-3] += idz*hdt*diss_top;
#endif

  /* Now do the implicit solving */
  /* First for U then for V */

  ok = tridiagonal(dde, lde, ude, rhse, dissd, work, N-2);
  if (ok != 0)
    error_message("** Critical Error:  Tridiagonal call for epsilon failed!");


  /* Next is tke boundary conditions */
  /* Bottom BC */

#ifdef EPSILON_DIRCHELET
  DR1(DISS,0) = diss_bed;
  DR1(DISS,N-1) = diss_top;
#endif
#ifdef EPSILON_FLUX
  DR1(DISS,0) = diss_bed;
  DR1(DISS,N-1) = DR1(DISS,N-2) + diss_top*dz/(kv*iSige);
#endif

  check_dissipation_stratified(DISS, N2);

}




void implicit_solve_kepsilonAB2(field1D *B, const field1D *Bt1, const field1D *Bt2, 
                                field1D *DISS, const field1D *DISSt1, const field1D *DISSt2,
                                const field1D *Kv, double k_bed, double diss_bed, double dt)
{

  int i, ok;
  int N = B->N;
  double hdt = 0.5*dt;

  static const double Sige =   1.077498753010693;
  static const double iSige =   9.280753200000001e-01;   //   this is 1/\sigma_\epsilon in GOTM
  static double cube_Cmu0 =   1.720651643280000e-01;

  double *bd = B->data+1;
  double *dissd = DISS->data+1;

  double *diffd = TKEDIFF->data+1;

  double bb, bp;
  double dd, dp;
  double qq;
  
  const double *bt1d = Bt1->data;
  const double *bt2d = Bt2->data;
  const double *dt1d = DISSt1->data;  
  const double *dt2d = DISSt2->data;  
  const double *kvd = Kv->data;

  double kv,kvm;
  double Ftke_p, Ftke_m, Fepsilon_m, Fepsilon_p, tmp;
  double tke_diffusion, ediffusion;
  double diss_top;

  /* First calculate the RHS stuff */
  /* do the first grid minus flux before looping */

  bb = *bd++;
  dd = *dissd++;
  kvm = *kvd++;
  Ftke_m = kvm * (bb - DR1(B,0));
  Fepsilon_m = iSige * kvm * (dd - DR1(DISS,0));

  for (i=1;i<N-1;i++) {

    bp = *bd++;
    dp = *dissd++;
    kv = *kvd++;

    Ftke_p = kv * ( bp - bb);
    Fepsilon_p = iSige * kv * (dp - dd);
    tke_diffusion = idz2*(Ftke_p-Ftke_m);
    ediffusion = idz2*(Fepsilon_p - Fepsilon_m);

    *diffd++ = tke_diffusion;
    
    rhsb[i-1] = bb + hdt*(tke_diffusion + 3.0* (*bt1d++) - *bt2d++);

    rhse[i-1] = dd + hdt*(ediffusion + 3.0* (*dt1d++) - *dt2d++);

    Ftke_m = Ftke_p;
    Fepsilon_m = Fepsilon_p;
    bb = bp;
    dd = dp;

  }

  /* i = 0 -> here we set prod=diss bottom bc */

  kvd = Kv->data;

  kvm = *kvd++;
  kv = *kvd++;
  ddb[0] = 1 + hdt*idz2*(kv+kvm);
  udb[0] = -hdt*kv*idz2;
  ldb[0] = 0.0;
  dde[0] = 1 + iSige*hdt*idz2*(kv+kvm);
  ude[0] = -hdt*iSige*kv*idz2;
  lde[0] = 0.0;

  /* What goes on rhs of rhb[0] for this??? */

  rhsb[0] += hdt*idz2*kvm*k_bed;
  rhse[0] += hdt*idz2*kvm*iSige*diss_bed;

  kvm = kv;
  /* i=1 to < N-3 */

  for (i=1;i<N-3;i++) {
    kv = *kvd++;
    tmp = hdt*idz2*(kvm+kv);
    ddb[i] = 1 + tmp;
    dde[i] = 1 + iSige*tmp;
    udb[i] = -hdt*kv*idz2;
    ude[i] = -iSige*hdt*kv*idz2;
    ldb[i] = -hdt*kvm*idz2;
    lde[i] = -iSige*hdt*kvm*idz2;
    kvm = kv;
  }

  /* i = N-3, here deal with top flux BC */

  i = N-3;
  kv = *kvd++;
  ldb[i] = -hdt*kvm*idz2;
  udb[i] = 0.0;
  ddb[i] = 1.0 + hdt*idz2*(kvm);
  lde[i] = -iSige*hdt*kvm*idz2;
  ude[i] = 0.0;

#ifdef EPSILON_DIRCHELET
  dde[i] = 1.0 + iSige*hdt*idz2*(kvm+kv);   // this is a Dirchelet BC!
#endif
#ifdef EPSILON_FLUX
  dde[i] = 1.0 + iSige*hdt*idz2*(kvm);   // this is a FLUX BC!
#endif

  /* Now handle the top tke boundary condition -> surface wave flux to rhs */

  rhsb[i] += idz*hdt*(wave_tke_surface_flux);

  /* Now do the implicit solving */
  /* First for U then for V */

  bd = B->data+1;   // reset the tke pointer


  ok = tridiagonal(ddb, ldb, udb, rhsb, bd, work, N-2);
  if (ok != 0)
    error_message("** Critical Error:  Tridiagonal call for k (in k-e) failed!");


  /* Next is tke boundary conditions */
  /* Bottom BC */
  DR1(B,0) = k_bed;

  /* Surface BC */
  kv = DR1(Kv,N-2);
  DR1(B,N-1) = DR1(B,N-2) + wave_tke_surface_flux*dz/kv;

  check_tke(B);

  /* figure out what the top dissipation should be */

  /* Next to the epsilon part */

  dissd = DISS->data+1;
  qq = sqrt(DR1(B,N-1));


#ifdef EPSILON_DIRCHELET
  diss_top  = cube_Cmu0 * CUBE(qq) / DR1(LL,N-1);
  rhse[N-3] += idz2*hdt*kv*diss_top;
#endif
#ifdef EPSILON_FLUX
  diss_top  = -cube_Cmu0 * kv * iSige * CUBE(qq) / (DR1(LL,N-1) * 0.2);
  rhse[N-3] += idz*hdt*diss_top;
#endif

  /* Now do the implicit solving */
  /* First for U then for V */

  ok = tridiagonal(dde, lde, ude, rhse, dissd, work, N-2);
  if (ok != 0)
    error_message("** Critical Error:  Tridiagonal call for epsilon failed!");


  /* Next is tke boundary conditions */
  /* Bottom BC */

#ifdef EPSILON_DIRCHELET
  DR1(DISS,0) = diss_bed;
  DR1(DISS,N-1) = diss_top;
#endif
#ifdef EPSILON_FLUX
  DR1(DISS,0) = diss_bed;
  DR1(DISS,N-1) = DR1(DISS,N-2) + diss_top*dz/(kv*iSige);
#endif


  check_dissipation_stratified(DISS, N2);
}


/* --------------------------------------------------
   ---------- Bubble implicit solvers ---------------
   --------------------------------------------------  */


void implicit_solve_bubble_euler(field2D *NB, const field2D *NBt, const field1D *Kv, double dt)
{

  int i,j, ok;
  int N = NB->N;
  int M = NB->M;
  double hdt = 0.5*dt;

  double *bd = NB->data+1;
  double bb, bp;
  
  field1D *B;   // this is WRONG!!
  const double *nbtd = NBt->data;
  
  const double *kvd = Kv->data;

  double *nbdp = NB->data+2*M;
  double *nbd = NB->data+M;
  double *nbdm = NB->data+M;
  double *rhsbd = RHS_BUBBLE->data;

  double *nb1d, *nb2d, *nb3d, *nb4d;
  double *radiusd;

  double kv,kvm;
  double Fdn, Fup, tmp, totFup;
  double bdiffusion;
  double rr;

  /* First calculate the RHS stuff */
  /* do the first grid minus flux before looping */

  for (i=1;i<N-1;i++) {
    totFup = 0.0;
    radiusd = Radius->data;
    kv = DR1(Kv,i);
    kvm = DR1(Kv,i-1);

    for (j=0;j<M;j++) {
  
      rr = *radiusd++;

      /* this is the bubble diffusion stuff */

      Fup = kv *(*nbdp++ - *nbd);
      Fdn = kvm *(*nbd - *nbdm++);
      totFup += CUBE(rr)*Fup;

      bdiffusion = idz2*(Fup-Fdn);

      *rhsbd++ = *nbd++ + dt*(0.5*bdiffusion + *nbtd++);  

    }
  }

  /* Now here the tridiagonal columns are set up */

  kvd = Kv->data;

  kvm = *kvd++;
  kv = *kvd++;
  ddb[0] = 1 + hdt*idz2*(kv);    // no diffusive flux near bed
  udb[0] = -hdt*kv*idz2;
  ldb[0] = 0.0;

  kvm = kv;

  /* i=1 to < N-3 */

  for (i=1;i<N-3;i++) {
    kv = *kvd++;
    tmp = 1 + hdt*idz2*(kvm+kv);
    ddb[i] = tmp;
    udb[i] = -hdt*kv*idz2;
    ldb[i] = -hdt*kvm*idz2;
    kvm = kv;
  }

  /* i = N-3, remember bubbles are NO diffusive flux */

  i = N-3;
  kv = *kvd++;
  ldb[i] = -hdt*kvm*idz2;
  udb[i] = 0.0;
  ddb[i] = 1.0 + hdt*idz2*(kvm);  // no diffusive flux @ surface

  /* Now do the implicit solving for the bubble */


  ok = tridiagonal_bubble2(ddb, ldb, udb, RHS_BUBBLE, XX_BUBBLE, work, work2);
  if (ok != 0)
    error_message("** Critical Error:  Tridiagonal call for bubble failed!");

  /* Now re-order into NB */

  nb1d = NB->data;
  nb2d = NB->data+M;

  nb3d = NB->row[N-2];
  nb4d = NB->row[N-1];

  for (j=0;j<M;j++) {
    *nb1d++ = *nb2d++;
    *nb3d++ = *nb4d++;
  }

}




void implicit_solve_bubbleAB2(field2D *NB, const field2D *NBt1, const field2D *NBt2, const field1D *Kv, double dt)
{

  int i,j, ok;
  int N = NB->N;
  int M = NB->M;
  double hdt = 0.5*dt;

  double *bd = NB->data+1;
  double bb, bp;
  
  field1D *B;   // this is WRONG!!
  const double *nbt1d = NBt1->data;
  const double *nbt2d = NBt2->data;
  
  const double *kvd = Kv->data;

  double *nbdp = NB->data+2*M;
  double *nbd = NB->data+M;
  double *nbdm = NB->data+M;
  double *rhsbd = RHS_BUBBLE->data;

  double *nb1d, *nb2d, *nb3d, *nb4d;
  double *radiusd;

  double kv,kvm;
  double Fdn, Fup, tmp, totFup;
  double bdiffusion;
  double rr;

  /* First calculate the RHS stuff */
  /* do the first grid minus flux before looping */

  for (i=1;i<N-1;i++) {
    totFup = 0.0;
    radiusd = Radius->data;
    kv = DR1(Kv,i);
    kvm = DR1(Kv,i-1);

    for (j=0;j<M;j++) {
  
      rr = *radiusd++;

      /* this is the bubble diffusion stuff */

      Fup = kv *(*nbdp++ - *nbd);
      Fdn = kvm *(*nbd - *nbdm++);
      totFup += CUBE(rr)*Fup;

      bdiffusion = idz2*(Fup-Fdn);

      *rhsbd++ = *nbd++ + hdt*(bdiffusion + 3.0 * (*nbt1d++) - *nbt2d++ );  

    }
  }

  /* Now here the tridiagonal columns are set up */

  kvd = Kv->data;

  kvm = *kvd++;
  kv = *kvd++;
  ddb[0] = 1 + hdt*idz2*(kv);    // no diffusive flux near bed
  udb[0] = -hdt*kv*idz2;
  ldb[0] = 0.0;

  kvm = kv;

  /* i=1 to < N-3 */

  for (i=1;i<N-3;i++) {
    kv = *kvd++;
    tmp = 1 + hdt*idz2*(kvm+kv);
    ddb[i] = tmp;
    udb[i] = -hdt*kv*idz2;
    ldb[i] = -hdt*kvm*idz2;
    kvm = kv;
  }

  /* i = N-3, remember bubbles are NO diffusive flux */

  i = N-3;
  kv = *kvd++;
  ldb[i] = -hdt*kvm*idz2;
  udb[i] = 0.0;
  ddb[i] = 1.0 + hdt*idz2*(kvm);  // no diffusive flux @ surface

  /* Now do the implicit solving for the bubble */


  ok = tridiagonal_bubble2(ddb, ldb, udb, RHS_BUBBLE, XX_BUBBLE, work, work2);
  if (ok != 0)
    error_message("** Critical Error:  Tridiagonal call for bubble failed!");

  /* Now re-order into NB */

  nb1d = NB->data;
  nb2d = NB->data+M;

  nb3d = NB->row[N-2];
  nb4d = NB->row[N-1];

  for (j=0;j<M;j++) {
    *nb1d++ = *nb2d++;
    *nb3d++ = *nb4d++;
  }

}
