/*
 * Copyright (c) 2002 Falk Feddersen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/* 
   implicit.h - definitions header file for implicit.c


*/

#ifndef IMPLICIT_H
#define IMPLICIT_H

#include "defs.h"

void init_implicit_solver(int nz, int nr, switch_t bubble_flag);

void deallocate_implicit_solver();

void implicit_solve_mflow_euler(field1D *U, const field1D *Ut, field1D *V, 
                                const field1D *Vt, const field1D *Kv, 
                                double dt);

void implicit_solve_mflowAB2(field1D *U, const field1D *Ut1, const field1D *Ut2,
                             field1D *V, const field1D *Vt1, const field1D *Vt2,
                             const field1D *Kv, double dt);

void implicit_solve_tke_euler(field1D *B, const field1D *Bt, const field1D *Kv, 
                              double k_bed, double dt);

void implicit_solve_tkeAB2(field1D *B, const field1D *Bt1, const field1D *Bt2,
                           const field1D *Kv, double k_bed, double dt);

void implicit_solve_kepsilon_euler(field1D *B, const field1D *Bt, field1D *DISS, const field1D *DISSt,
				   const field1D *Kv, double k_bed, double diss_bed, double dt);

void implicit_solve_kepsilonAB2(field1D *B, const field1D *Bt1, const field1D *Bt2, 
                                field1D *DISS, const field1D *DISSt1, const field1D *DISSt2,
	  		        const field1D *Kv, double k_bed, double diss_bed, double dt);


void implicit_solve_bubble_euler(field2D *NB, const field2D *NBt, const field1D *Kv, double dt);

void implicit_solve_bubbleAB2(field2D *NB, const field2D *NBt1, const field2D *NBt2, const field1D *Kv, double dt);


#endif /* IMPLICIT_H */
