/*
 * Copyright (c) 2002 Falk Feddersen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


/* --  pinput.h: loads and processes the init file */


#ifndef PINPUT_H
#define PINPUT_H


#include <glib.h>

#include "timing.h"
#include "outputflags.h"
#include "wave.h"
#include "mflow.h"
#include "bubble.h"
#include "sediment.h"
#include "density.h"

/* Timing routines and data structures */
         
typedef enum {
  KEPSILON, MELLOR_YAMADA, CRAIG_BANNER
} model_t;



typedef struct {
  GString *input_file;
  int nz;
  gdouble dz;              // this must be stored in units of meters!
  gdouble depth;
  gdouble z0b;
  gdouble z0s;
  int tke_ic; 
  GString *tke_ic_file; 
  int tke_bottom_bc;   
  outputflags *A;
  timing *T;
  bubbleinfo *B;      // this should take over from a bunch of stuff
  waveinfo *W;
  mflowinfo *M;
  sedimentinfo *SED;
  densityinfo *D;
  model_t model_type;
} inputinfo;


/* timing stuff */

void timing_report(timing *T);

int fix_units(char *unit);

gdouble change_units(gdouble t, int u);

gdouble get_time_in_current_units(timing *T);

gdouble fmodfalk(gdouble num, gdouble denom);

//timing *process_timing(FILE *fp);

inputinfo* load_init_file(GString *fname);

void process_init_file(inputinfo *I);

void outputflags_destroy(outputflags *A);

#endif    /* PINPUT_H */
