
disp('Plunging Wave Test')

nz = 401;

tke = load_binary_file('tke_pt.dat',nz);
diss = load_binary_file('diss_pt.dat',nz);
Kv = load_binary_file('KUV_pt.dat',nz-1);
tkediff = load_binary_file('tkediff_pt.dat',nz);

[n,m] = size(tke);

dz = 0.01;


mdiss = mean(diss);
mtkediff = mean(tkediff);


z = [0:m-1]*dz;
z2 = ([0:m-2]+0.5)*dz;
z3 = [1:m-2]*dz;

bottom = 0.01;
top = z(m);

dt = 0.1;
z0b = 0.005;

Cmu0 = 0.5562;

L = Cmu0^3 * tke.^(3/2) ./ diss;
Lm = 0.4*(z+z0b);

time = [0:n-1]*dt;


% Integrate the dissipation 

imdiss = dz*0.5*sum(mdiss(1:nz-1) + mdiss(2:nz))

iepsilon = 0.02;


figure(1),clf, orient tall

subplot(4,2,1)
semilogx(tke(n-80:2:n,:),z),grid
axis([0.8*min(min(tke(n-80:2:n,:))) 1.2*max(max(tke(n-80:2:n,:)))  bottom top])
xlabel('tke (m^2/s^2)')
ylabel('Depth (m)')

subplot(4,2,2)
semilogx(diss(n-80:2:n,:),z)  %,mdiss,z,'--',abs(tkediff(n-80:2:n,:)),z,':')
grid
axis([0.8*min(min(diss(n-80:2:n,:))) 1.2*max(max(diss(n-80:2:n,:)))  bottom top])
%legend('model \epsilon','theory \epsilon','production','tke diff')
xlabel('\epsilon (m^2/s^3)')
ylabel('Depth (m)')


subplot(4,2,5)
semilogx(Kv(n-80:2:n,:),z2)   %,KQV(n-80:2:n,:),z2,'--',mvt,z2,mvt2,z,'-.')
%axis([0 0.01 bottom top])
%legend('K_{uv}','K_{qv}','theory','k-e',4)
xlabel('K_v (m^2/s)')
ylabel('Depth (m)')

subplot(4,2,6)
loglog(Kv(n-80:2:n,:),z2)    %,KQV(n-80:2:n,:),z2,'--',mvt,z2,mvt2,z,'-.')
axis([1e-4 1e-1 bottom top])
xlabel('K_v (m^2/s)')
ylabel('Depth (m)')

subplot(4,2,7)
plot(L,z,Lm,z,'--')
xlabel('Length Scale (C_\mu)^3 k^{3/2}/\epsilon (m),  kappa*(z+z0)=dashed')
ylabel('Depth (m)')


subplot(4,2,8)
semilogx(mdiss,z,abs(mtkediff),z,'-.')
axis([1e-7 10 bottom top])
legend('\epsilon','tke diff')
xlabel('Mean TKE Dynamics')
ylabel('Depth (m)')
   
