
load -ascii V.dat
load -ascii U.dat
load -ascii VW.dat
load -ascii tke.dat
load -ascii diss.dat
load -ascii KUV.dat
load -ascii KQV.dat
load -ascii tottke.dat
load -ascii PROD.dat
load -ascii tkediff.dat

[n,m] = size(V);

nz = m;
dz = 0.01;


z = [0:m-1]*dz;
z2 = ([0:m-2]+0.5)*dz;
z3 = [1:m-2]*dz;

bottom = 0.01;
top = z(m);

dt = 120;
vstar = 0.0153;
tauy = vstar*vstar;
z0b = 0.005;

time = [0:n-1]*dt;

% log fit of V 

VM = mean(V(n-159:n,:));

ifit = [20:50];


vb = VM(ifit)';
zb = z(ifit)';
lzb = log(zb);

CC = corrcoef(zb,vb);
r = CC(1,2);

[nb,mb] = size(vb);

G = [lzb ones(nb,1)];

mm7 = inv(G'*G)*G'*vb;

xx = [-3.5 -0.5];
yy = mm7(1)*xx + mm7(2);

vstar_m = mm7(1)*0.4;
z0b_m = exp(-mm7(2)/mm7(1));

stest = sprintf('True vstar = %g, log fit vstar=%.3g,  True z0b = %g Log fit z0b = %.3g',vstar,vstar_m,z0b, z0b_m);
disp(stest);

mvt = 0.4*vstar*(z2+z0b);   % constant stress eddy viscosity

Cmu0 = 0.5562;
Ce1 = 1.44;
Ce2 = 1.92;
Sige = 1.08;

mk = (vstar/Cmu0)^2 * ones(m,1);     % model kinetic energy
 
mdiss = vstar^3./(0.4*(z+z0b));   % model dissipation

mvt2 = Cmu0^4 * tke(n,:).^2 ./ diss(n,:);  % k-e model Kv

mV = (vstar/0.4) * log((z+z0b)/z0b);


L = Cmu0^3 * tke(n,:).^(3/2) ./ diss(n,:);
Lm = 0.4*(z+z0b);

eprod = Ce1 * PROD(n,:) .* diss(n,:) ./ tke(n,:);
ediss = Ce2 * diss(n,:) .* diss(n,:) ./ tke(n,:);
ediff = diff(KUV(n,:) .* diff(diss(n,:))/dz)/(Sige*dz);

eerr = eprod(2:m-1) - ediss(2:m-1) + ediff;
mean(eerr)
std(eerr)

figure(1),clf, orient tall

subplot(4,2,1)
semilogy(time,tottke,time,V(:,10),'--',time,diss(:,10),':')
grid
legend('tke','V','diss',4)
xlabel('Time (sec)')
ylabel('tke, U, & V')


subplot(4,2,2)
plot(U(n-16:2:n,:),z)
xlabel('U & V (m/s)')
ylabel('Depth (z)')
%axis([-0.1 0.3 bottom top])

subplot(4,2,3)
plot(V(n-16:2:n,:),z),grid
xlabel('V (m/s)')
ylabel('Depth (z)')
%axis([-0.1 0.3 bottom  top])

subplot(4,2,4)
semilogy(tke(n-16:2:n,:),z),grid
%axis([0 1e-3 bottom top])
xlabel('tke (m^2/s^2)')
ylabel('Depth (m)')

subplot(4,2,5)
semilogy(VW(n-16:2:n,:),z2)
xlabel('<vw> (m^2/s^2)')
ylabel('Depth (m)')
%axis([0 3e-4 bottom top])


subplot(4,2,6)
loglog(diss(n-16:2:n,:),z)  %,mdiss,z,'--',PROD(n-16:2:n,:),z,':',abs(tkediff(n-16:2:n,:)),z,'-.')
%axis([0 8e-4 bottom top])
%legend('model \epsilon','theory \epsilon','production','tke diff')
xlabel('\epsilon (m^2/s^3)')
ylabel('Depth (m)')


subplot(4,2,7)
semilogy(KUV(n-16:2:n,:),z2)   %,KQV(n-16:2:n,:),z2,'--',mvt,z2,mvt2,z,'-.')
%axis([0 0.01 bottom top])
%legend('K_{uv}','K_{qv}','theory','k-e',4)
xlabel('K_v (m^2/s)')
ylabel('Depth (m)')

subplot(4,2,8)
loglog(KUV(n-16:2:n,:),z2)    %,KQV(n-16:2:n,:),z2,'--',mvt,z2,mvt2,z,'-.')
%axis([0 0.01 bottom top])
%legend('K_{uv}','K_{qv}','theory','k-e',2)
xlabel('K_v (m^2/s)')
ylabel('Depth (m)')



figure(2),clf, orient tall

subplot(3,1,1)
semilogx(eprod,z,ediss,z,'--',ediff,z3,':')
legend('\epsilon prod','\epsilon diss','\epsilon diff')
xlabel('\epsilon dynamics (m^2/s^4)')
ylabel('Depth (m)')

subplot(3,1,2)
plot(L,z,Lm,z,'--')
legend('model L','theory L',4)
xlabel('Length Scale (C_\mu)^3 k^{3/2}/\epsilon (m)')
ylabel('Depth (m)')


if (max(abs(eerr)>1e-8))
   disp('*** ERROR:  error in epsilon eq too large')
end;
   

% Boundary Conditions 

ee1 = Cmu0^3 * tke(n,1)^(3/2)/(0.4*z0b)

abs(diss(n,1)-ee1)
