/*
 * Copyright (c) 2001 Falk Feddersen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


/* -- timing.c: handles timing input for model */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "timing.h"

void parse_error_message(char *s);


char *units[4]={"(sec)","(min)","(hr)","(day)"};

/* returns index for which units are represented by the input string */

int fix_units(char *unit)
{
  int i;

  for (i=0;i<4;i++)
    if (strcmp(unit,units[i])==0)
      return i;
  
  fprintf(stderr,"Syntax Error: Don't recognize the time units %s.\n",unit);
  exit(-1);
}

/* Convert the units into seconds - the default units */

gdouble change_units(gdouble t, int u)
{
  switch (u) {
  case 0: return t;
  case 1: return 60.0*t;
  case 2: return 3600.0*t;
  case 3: return 24.0*3600.0*t;
  default: parse_error_message("Bad flag to change_units!");
  }

  return -1.0;   // this should never happen!
}

gdouble get_time_in_current_units(timing *T)
{
  gdouble time;

  time = T->time_index * T->dt;

  switch (T->units_flag) {
  case 0: return time;
  case 1: return time/60.0;
  case 2: return time/3600.0;
  case 3: return time/(24.0*3600.0);
  default: parse_error_message("Bad flag to change_units_back!");
  }

  return -1.0;  // this should never happen!
}


gdouble fmodfalk(gdouble num, gdouble denom)
{

   gdouble t1;
   gdouble t2=0.0, t3=0.0;

   t1 = num/denom;
   t2 = modf(t1,&t3);

   return t2;

}

int fdivfalk(gdouble num, gdouble denom)
{

   gdouble t1;
   gdouble t2=0.0, t3=0.0;
   int tt3;

   t1 = num/denom;
   t2 = modf(t1,&t3);

   tt3 = (int ) t3;

   return tt3;

}


timing* process_timing(char *s)
{

  gdouble time1,time2,time3,dt;
  gdouble t1, t2, t3, dt2;
  int num;

  char units1[12], units2[12], units3[12], units4[12];
  char tstr[120];
  int un1, un2, un3, un4;

  timing *T;

  /* allocate the timing */

  T = (timing * ) g_malloc(sizeof(timing));

  T->time_index = 0;
  T->finished = 0;

  /* load individual component pieces */

  if ((num=sscanf(s,"%s %lf  %s  %lf %s %lf %s %lf %s",tstr,&t1,units1,&t2,units2,&t3,units3,&dt2, units4))!=9) 
    parse_error_message("Syntax Error: Timing information incorrect - expecting 'timing ... ' - please check.");

  if (strcasecmp(tstr,"timing")) 
    parse_error_message("Syntax Error: expecting 'timing .....'  - please check.");

  un1 = fix_units(units1);        un2 = fix_units(units2);
  un3 = fix_units(units3);        un4 = fix_units(units4);

  T->units_ptr = units[un1];
  T->units_flag = un1;
 
  time1 = change_units(t1,un1);     time3 = change_units(t3,un3);
  time2 = change_units(t2,un2);     dt = change_units(dt2,un4);

  /* check for AB3 time stepping */

#ifdef AB3
     dt2 = 3.0*dt;
#else
     dt2 = 2.0*dt;
#endif

   if (fmodfalk(time1,time2)!=0.0) 
     parse_error_message("Syntax Error: Time1 not integral to Time2.  Check init file.");
   else 
     T->level3_iter = fdivfalk(time1,time2);

   if (fmodfalk(time2,time3)!=0.0) 
     parse_error_message("Syntax Error: Time2 not integral to Time3.  Check init file.");
   else 
     T->level2_iter = fdivfalk(time2,time3);
   if (fmodfalk(time3,dt2)!=0.0) 
     parse_error_message("Syntax Error: Time3 not integral to dt.  Check init file.");
   else 
     T->level1_iter = fdivfalk(time3,dt2);

   T->total_run_time = time1;
   T->level1_time = time3;
   T->level2_time = time2;
   T->dt = dt;

   return T;
}
